import os from 'node:os';
import { execFile } from 'node:child_process';
import path from 'node:path';
import { createHash, randomUUID } from 'node:crypto';
import fs from 'node:fs';

function mapArchAndPlatform(platform, arch, ver = "") {
	if (platform === "linux") {
		switch (arch) {
			case "x64":
				return {file: "linux-x64", link: `node-${ver}-linux-x64`}

			case "arm64":
				return {file: "linux-arm64", link: `node-${ver}-linux-arm64`}
		}
	} else if (platform === "darwin") {
		switch (arch) {
			case "x64":
				return {file: "osx-x64-tar", link: `node-${ver}-darwin-x64`}

			case "arm64":
				return {file: "osx-arm64-tar", link: `node-${ver}-darwin-arm64`}
		}
	}

	throw new Error(`Unsupported platform/arch combination.`)
}

async function impl_getAvailableVersions() {
	let ret = [];

	const arch_platform_identifier = mapArchAndPlatform(os.platform(), os.arch()).file;

	const response = await fetch("https://nodejs.org/dist/index.tab");
	const list = await response.text();
	// first line is description
	const entries = list.split("\n").slice(1);

	for (const entry of entries) {
		const fields = entry.split("\t");

		if (2 >= fields.length) continue

		const [version, date, files_str] = fields;
		const files = files_str.split(",");

		if (!files.includes(arch_platform_identifier)) continue

		ret.push(version);
	}

	return ret
}

async function impl_getDownloadURL(version) {
	const arch_platform_identifier = mapArchAndPlatform(
		os.platform(), os.arch(), version
	).link;

	return `https://nodejs.org/dist/${version}/${arch_platform_identifier}.tar.gz`
}

function nodeFsReadFileInChunksSync(file_path, chunk_size) {
	const fd = fs.openSync(file_path, "r");
	let closed = false;

	return {
		readNextChunk() {
			if (closed) {
				throw new Error(`Cannot call readNextChunk() on closed instance.`)
			}

			let chunk_buffer = Buffer.alloc(chunk_size);

			const bytes_read = fs.readSync(
				fd, chunk_buffer
			);

			if (!bytes_read) {
				if (!closed) fs.closeSync(fd);

				closed = true;

				return false
			}

			return chunk_buffer.subarray(0, bytes_read)
		},

		close() {
			if (closed) return

			fs.closeSync(fd);
			closed = true;
		}
	}
}

function index$3(file_path, algo, hash_encoding = "hex") {
	const hash = createHash(algo);

	//
	// read file in 4096 byte chunks
	//
	const file = nodeFsReadFileInChunksSync(file_path, 4096);

	while (true) {
		const chunk = file.readNextChunk();

		if (!chunk) break

		hash.update(chunk);
	}

	hash.setEncoding(hash_encoding);
	hash.end();

	return hash.read()
}

function tryStat$1(path, fn = "lstat") {
	try {
		const stat = fs[`${fn}Sync`](path);

		return stat
	} catch (error) {
		//
		// ENOENT means no such entry
		//
		if (error.code === "ENOENT") {
			return false
		}

		// re-throw every other error
		throw error
	}
}

function fsGetPathTypeSync$1(...args) {
	const path_to_check = path.join(...args);

	const lstat = tryStat$1(path_to_check);

	if (lstat === false) return false

	if (lstat.isSymbolicLink()) {
		const stat = tryStat$1(path_to_check, "stat");

		if (stat === false) return "link->broken"
		if (stat.isDirectory()) return "link->dir"

		return "link->file"
	}

	if (lstat.isDirectory()) return "dir"

	return "file"
}

function deleteSymbolicLink(path) {
	fs.unlinkSync(path);
}

function deleteFile(path) {
	fs.unlinkSync(path);
}

function _scandir(context, relative_entry_dir, callback) {
	const entries = fs.readdirSync(
		path.resolve(context.root, relative_entry_dir)
	);

	for (const entry of entries) {
		const absolute_path = path.join(
			context.root, relative_entry_dir, entry
		);

		const relative_path = path.join(
			relative_entry_dir, entry
		);

		const stats = fs.lstatSync(absolute_path);

		let type = "file";

		if (stats.isSymbolicLink()) {
			type = "link";
		} else if (stats.isDirectory()) {
			type = "dir";
		}

		const handle_entry = () => {
			if (typeof callback === "function") {
				callback(type, {relative_path, absolute_path});
			}

			context.entries.push({type, relative_path, absolute_path});
		};

		if (type === "dir") {
			if (context.order) {
				_scandir(context, relative_path, callback);
			}

			handle_entry();

			if (!context.order) {
				_scandir(context, relative_path, callback);
			}
		} else {
			handle_entry();
		}
	}
}

function scandir(dir_path, callback, order) {
	let entries = [];

	_scandir({
		order,
		entries,
		root: fs.realpathSync(dir_path)
	}, ".", callback);

	return entries
}

function nodeFsScandirSync(dir_path, callback) {
	return scandir(dir_path, callback, false)
}

nodeFsScandirSync.reverse = function(dir_path, callback) {
	return scandir(dir_path, callback, true)
};

function deleteDirectory(path) {
	const entries = nodeFsScandirSync.reverse(path);

	for (const entry of entries) {
		if (entry.type === "link") {
			deleteSymbolicLink(entry.absolute_path);
		} else if (entry.type === "dir") {
			fs.rmdirSync(entry.absolute_path);
		} else {
			deleteFile(entry.absolute_path);
		}
	}

	fs.rmdirSync(path);
}

const remove_fn = {
	"link->dir": deleteSymbolicLink,
	"link->file": deleteSymbolicLink,
	"link->broken": deleteSymbolicLink,
	"file": deleteFile,
	"dir": deleteDirectory
};

function index$1(path) {
	const path_type = fsGetPathTypeSync$1(path);

	//
	// do nothing if path doesn't exist
	//
	if (path_type === false) {
		return
	} else if (!(path_type in remove_fn)) {
		throw new Error(`I don't know how to remove a path of type '${path_type}'.`)
	}

	const fn = remove_fn[path_type];

	return fn(path)
}

function hashFileSync(...args) {
	return index$3(...args)
}

function removeSync(path) {
	return index$1(path)
}

function generateTemporaryPathNameSync() {
	return path.join(os.tmpdir(), randomUUID())
}

//
// todo: replace with fetch()
//
function downloadFileViaCURL(url) {
	return new Promise((resolve, reject) => {
		const file_dest_path = generateTemporaryPathNameSync();

		execFile("curl", [
			url, "--output", file_dest_path
		], {
			stdio: "pipe"
		}, (error, stdout, stderr) => {
			if (error) {
				reject(error);
			} else {
				resolve(file_dest_path);
			}
		});
	})
}

function extractArchiveWithTar(tar_file) {
	return new Promise((resolve, reject) => {
		const dir_dest_path = generateTemporaryPathNameSync();

		fs.mkdirSync(dir_dest_path);

		execFile("tar", [
			"-xzf", tar_file, "-C", "."
		], {
			cwd: dir_dest_path,
			stdio: "pipe"
		}, (error, stdout, stderr) => {
			if (error) {
				reject(error);
			} else {
				resolve(dir_dest_path);
			}
		});
	})
}

async function getChecksums(version, arch_platform_identifier) {
	const response = await fetch(`https://nodejs.org/dist/${version}/SHASUMS256.txt`);
	const body = await response.text();
	const entries = body.split("\n");

	let target_checksum = "";

	for (const entry of entries) {
		const fields = entry.trim().split(" ").filter(p => p.length);

		if (fields.length !== 2) continue

		const [checksum, file_name] = fields;

		if (file_name === `${arch_platform_identifier}.tar.gz`) {
			target_checksum = checksum;

			break
		}
	}

	if (target_checksum.length !== 64) {
		throw new Error(`Unable to get correct checksum for download.`)
	}

	return target_checksum
}

async function impl_download(version) {
	const versions = await impl_getAvailableVersions();

	if (!versions.includes(version)) {
		throw new Error(`No such version '${version}'.`)
	}

	const arch_platform_identifier = mapArchAndPlatform(
		os.platform(), os.arch(), version
	).link;

	const file_checksum = await getChecksums(version, arch_platform_identifier);

	const url = await impl_getDownloadURL(version);

	let temp_archive_file = null, temp_directory = null;

	try {
		temp_archive_file = await downloadFileViaCURL(url);

		const local_checksum = hashFileSync(temp_archive_file, "sha256");

		if (file_checksum !== local_checksum) {
			throw new Error(`Checksum error <${file_checksum} != ${local_checksum}>.`)
		}

		temp_directory = await extractArchiveWithTar(temp_archive_file);

		removeSync(temp_archive_file);

		return {
			path: path.join(temp_directory, arch_platform_identifier),
			cleanup() {
				removeSync(temp_directory);
			}
		}
	} catch (error) {
		//
		// remove temporary items
		//
		if (temp_archive_file !== null) removeSync(temp_archive_file);
		if (temp_directory !== null) removeSync(temp_directory);

		throw error
	}
}

const getAvailableVersions = impl_getAvailableVersions;
const getDownloadURL = impl_getDownloadURL;
const download = impl_download;

var index = {
	getAvailableVersions,
	getDownloadURL,
	download
};

export { index as default, download, getAvailableVersions, getDownloadURL };
